@@ -63,7 +63,8 @@
 
 
 
-#undef DUMP_IMAGE_DETAILS
+#undef DUMP_IMAGE_DETAILS 
+#undef IO_GIFDEBUG
 
 #define MAXCOLORMAPSIZE  256
 #define MAX_LZW_BITS     12
@@ -83,7 +84,7 @@ enum {
 	GIF_GET_COLORMAP,
 	GIF_GET_NEXT_STEP,
 	GIF_GET_FRAME_INFO,
-	GIF_GET_EXTENTION,
+	GIF_GET_EXTENSION,
 	GIF_GET_COLORMAP2,
 	GIF_PREPARE_LZW,
 	GIF_LZW_FILL_BUFFER,
@@ -116,6 +117,7 @@ struct _GifContext
         unsigned int global_bit_pixel;
 	unsigned int global_color_resolution;
         unsigned int background_index;
+        gboolean stop_after_first_frame;
 
         gboolean frame_cmap_active;
         CMap frame_color_map;
@@ -138,8 +140,8 @@ struct _GifContext
 	FILE *file;
 
 	/* progressive read, only. */
-	ModulePreparedNotifyFunc prepare_func;
-	ModuleUpdatedNotifyFunc update_func;
+	GdkPixbufModulePreparedFunc prepare_func;
+	GdkPixbufModuleUpdatedFunc update_func;
 	gpointer user_data;
         guchar *buf;
 	guint ptr;
@@ -149,6 +151,7 @@ struct _GifContext
 	/* extension context */
 	guchar extension_label;
 	guchar extension_flag;
+        gboolean in_loop_extension;
 
 	/* get block context */
 	guchar block_count;
@@ -211,11 +214,14 @@ gif_read (GifContext *context, guchar *buffer, size_t len)
 #endif
 		retval = (fread(buffer, len, 1, context->file) != 0);
 
-                if (!retval && ferror (context->file))
+                if (!retval && ferror (context->file)) {
+                        gint save_errno = errno;
                         g_set_error (context->error,
                                      G_FILE_ERROR,
-                                     g_file_error_from_errno (errno),
-                                     _("Failure reading GIF: %s"), strerror (errno));
+                                     g_file_error_from_errno (save_errno),
+                                     _("Failure reading GIF: %s"), 
+                                     strerror (save_errno));
+                }
                 
 #ifdef IO_GIFDEBUG
 		if (len < 100) {
@@ -249,7 +255,7 @@ gif_read (GifContext *context, guchar *buffer, size_t len)
 		}
 		context->amount_needed = len - (context->size - context->ptr);
 	}
-	return 0;
+	return FALSE;
 }
 
 /* Changes the stage to be GIF_GET_COLORMAP */
@@ -353,7 +359,7 @@ get_data_block (GifContext *context,
 static void
 gif_set_get_extension (GifContext *context)
 {
-	context->state = GIF_GET_EXTENTION;
+	context->state = GIF_GET_EXTENSION;
 	context->extension_flag = TRUE;
 	context->extension_label = 0;
 	context->block_count = 0;
@@ -376,8 +382,8 @@ gif_get_extension (GifContext *context)
 		}
 
 		switch (context->extension_label) {
-		case 0xf9:			/* Graphic Control Extension */
-			retval = get_data_block (context, (unsigned char *) context->block_buf, NULL);
+                case 0xf9:			/* Graphic Control Extension */
+                        retval = get_data_block (context, (unsigned char *) context->block_buf, NULL);
 			if (retval != 0)
 				return retval;
 
@@ -399,6 +405,36 @@ gif_get_extension (GifContext *context)
 			/* Now we've successfully loaded this one, we continue on our way */
 			context->block_count = 0;
 			context->extension_flag = FALSE;
+			break;
+                case 0xff: /* application extension */
+                        if (!context->in_loop_extension) { 
+                                retval = get_data_block (context, (unsigned char *) context->block_buf, NULL);
+                                if (retval != 0)
+                                        return retval;
+                                if (!strncmp (context->block_buf, "NETSCAPE2.0", 11) ||
+                                    !strncmp (context->block_buf, "ANIMEXTS1.0", 11)) {
+                                        context->in_loop_extension = TRUE;
+                                }
+                                context->block_count = 0;
+                        }
+                        if (context->in_loop_extension) {
+                                do {
+                                        retval = get_data_block (context, (unsigned char *) context->block_buf, &empty_block);
+                                        if (retval != 0)
+                                                return retval;
+                                        if (context->block_buf[0] == 0x01) {
+                                                context->animation->loop = context->block_buf[1] + (context->block_buf[2] << 8);
+                                                if (context->animation->loop != 0) 
+                                                        context->animation->loop++;
+                                        }
+                                        context->block_count = 0;
+                                }
+                                while (!empty_block);
+                                context->in_loop_extension = FALSE;
+                                context->extension_flag = FALSE;
+                                return 0;
+                        }
+			break;                          
 		default:
 			/* Unhandled extension */
 			break;
@@ -534,6 +570,17 @@ gif_lzw_clear_code (GifContext *context)
 	return 0;
 }
 
+#define CHECK_LZW_SP() G_STMT_START {                                           \
+        if ((guchar *)context->lzw_sp >=                                        \
+            (guchar *)context->lzw_stack + sizeof (context->lzw_stack)) {       \
+                 g_set_error (context->error,                                   \
+                             GDK_PIXBUF_ERROR,                                  \
+                             GDK_PIXBUF_ERROR_CORRUPT_IMAGE,                    \
+                             _("Stack overflow"));                              \
+                return -2;                                                      \
+        }                                                                       \
+} G_STMT_END
+
 static int
 lzw_read_byte (GifContext *context)
 {
@@ -589,7 +636,7 @@ lzw_read_byte (GifContext *context)
                         g_set_error (context->error,
                                      GDK_PIXBUF_ERROR,
                                      GDK_PIXBUF_ERROR_FAILED,
-                                     _("GIF image loader can't understand this image."));
+                                     _("GIF image loader cannot understand this image."));
                         return -2;
                         
 			if (ZeroDataBlock) {
@@ -608,19 +655,20 @@ lzw_read_byte (GifContext *context)
 		incode = code;
 
 		if (code >= context->lzw_max_code) {
+                        CHECK_LZW_SP ();
 			*(context->lzw_sp)++ = context->lzw_firstcode;
 			code = context->lzw_oldcode;
 		}
 
 		while (code >= context->lzw_clear_code) {
-                        if ((code >= (1 << MAX_LZW_BITS)) 
-                            || (context->lzw_sp >= context->lzw_stack + ((1 << (MAX_LZW_BITS)) * 2 + 1))) {
+                        if (code >= (1 << MAX_LZW_BITS)) {
                                 g_set_error (context->error,
                                              GDK_PIXBUF_ERROR,
                                              GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
                                              _("Bad code encountered"));
 				return -2;
                         }
+                        CHECK_LZW_SP ();
 			*(context->lzw_sp)++ = context->lzw_table[1][code];
 
 			if (code == context->lzw_table[0][code]) {
@@ -633,6 +681,7 @@ lzw_read_byte (GifContext *context)
 			code = context->lzw_table[0][code];
 		}
 
+                CHECK_LZW_SP ();
 		*(context->lzw_sp)++ = context->lzw_firstcode = context->lzw_table[1][code];
 
 		if ((code = context->lzw_max_code) < (1 << MAX_LZW_BITS)) {
@@ -725,13 +774,49 @@ gif_fill_in_lines (GifContext *context, guchar *dest, guchar v)
 	}
 }
 
-static void
-set_need_recomposite (gpointer data, gpointer user_data)
+/* Clips a rectancle to the base dimensions. Returns TRUE if the clipped rectangle is non-empty. */
+static gboolean
+clip_frame (GifContext *context, 
+            gint       *x, 
+            gint       *y, 
+            gint       *width, 
+            gint       *height)
 {
-        GdkPixbufFrame *frame = (GdkPixbufFrame *)data;
-        frame->need_recomposite = TRUE;
+        gint orig_x, orig_y;
+        
+        orig_x = *x;
+        orig_y = *y;
+	*x = MAX (0, *x);
+	*y = MAX (0, *y);
+	*width = MIN (context->width, orig_x + *width) - *x;
+	*height = MIN (context->height, orig_y + *height) - *y;
+
+	if (*width > 0 && *height > 0)
+		return TRUE;
+
+	/* The frame is completely off-bounds */
+
+	*x = 0;
+	*y = 0;
+	*width = 0;
+	*height = 0;
+
+        return FALSE;
 }
 
+/* Call update_func on the given rectangle, unless it is completely off-bounds */
+static void
+maybe_update (GifContext *context,
+              gint        x,
+              gint        y,
+              gint        width,
+              gint        height)
+{
+        if (clip_frame (context, &x, &y, &width, &height))
+                (*context->update_func) (context->frame->pixbuf, 
+                                         x, y, width, height,
+                                         context->user_data);
+}
 
 static int
 gif_get_lzw (GifContext *context)
@@ -748,12 +833,31 @@ gif_get_lzw (GifContext *context)
                 context->frame->composited = NULL;
                 context->frame->revert = NULL;
                 
-                context->frame->pixbuf =
-                        gdk_pixbuf_new (GDK_COLORSPACE_RGB,
-                                        TRUE,
-                                        8,
-                                        context->frame_len,
-                                        context->frame_height);
+                if (context->frame_len == 0 || context->frame_height == 0) {
+                        /* An empty frame, we just output a single transparent
+                         * pixel at (0, 0).
+                         */
+                        context->x_offset = 0;
+                        context->y_offset = 0;
+                        context->frame_len = 1;
+                        context->frame_height = 1;
+                        context->frame->pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8, 1, 1);
+                        if (context->frame->pixbuf) {
+                                guchar *pixels;
+
+                                pixels = gdk_pixbuf_get_pixels (context->frame->pixbuf);
+                                pixels[0] = 0;
+                                pixels[1] = 0;
+                                pixels[2] = 0;
+                                pixels[3] = 0;
+                        }
+                } else
+                        context->frame->pixbuf =
+                                gdk_pixbuf_new (GDK_COLORSPACE_RGB,
+                                                TRUE,
+                                                8,
+                                                context->frame_len,
+                                                context->frame_height);
                 if (!context->frame->pixbuf) {
                         g_free (context->frame);
                         g_set_error (context->error,
@@ -799,25 +903,8 @@ gif_get_lzw (GifContext *context)
 
                 context->frame->bg_transparent = (context->gif89.transparent == context->background_index);
                 
-                {
-                        /* Update animation size */
-                        int w, h;
-                        
-                        context->animation->n_frames ++;
-                        context->animation->frames = g_list_append (context->animation->frames, context->frame);
-
-                        w = context->frame->x_offset +
-                                gdk_pixbuf_get_width (context->frame->pixbuf);
-                        h = context->frame->y_offset +
-                                gdk_pixbuf_get_height (context->frame->pixbuf);
-                        if (w  > context->animation->width || h > context->animation->height) {
-                                g_list_foreach (context->animation->frames, set_need_recomposite, NULL);
-                        }
-                        if (w > context->animation->width)
-                                context->animation->width = w;
-                        if (h > context->animation->height)
-                                context->animation->height = h;
-                }
+                context->animation->n_frames ++;
+                context->animation->frames = g_list_append (context->animation->frames, context->frame);
 
                 /* Only call prepare_func for the first frame */
 		if (context->animation->frames->next == NULL) { 
@@ -829,6 +916,7 @@ gif_get_lzw (GifContext *context)
                         /* Otherwise init frame with last frame */
                         GList *link;
                         GdkPixbufFrame *prev_frame;
+                        gint x, y, w, h;
                         
                         link = g_list_find (context->animation->frames, context->frame);
 
@@ -836,13 +924,43 @@ gif_get_lzw (GifContext *context)
 
                         gdk_pixbuf_gif_anim_frame_composite (context->animation, prev_frame);
 
-                        gdk_pixbuf_copy_area (prev_frame->composited,
-                                              context->frame->x_offset,
-                                              context->frame->y_offset,
-                                              gdk_pixbuf_get_width (context->frame->pixbuf),
-                                              gdk_pixbuf_get_height (context->frame->pixbuf),
-                                              context->frame->pixbuf,
-                                              0, 0);
+                        /* Composite failed */
+                        if (prev_frame->composited == NULL) {
+                                GdkPixbufFrame *frame = NULL;
+                                link = g_list_first (context->animation->frames);
+                                while (link != NULL) {
+                                        frame = (GdkPixbufFrame *)link->data;
+                                        if (frame != NULL) {
+                                                if (frame->pixbuf != NULL)
+                                                        g_object_unref (frame->pixbuf);
+                                                if (frame->composited != NULL)
+                                                        g_object_unref (frame->composited);
+                                                if (frame->revert != NULL)
+                                                        g_object_unref (frame->revert);
+                                                g_free (frame);
+                                        }
+                                        link = link->next;
+                                }
+                                
+                                g_list_free (context->animation->frames);
+                                context->animation->frames = NULL;
+                                
+                                g_set_error (context->error,
+                                             GDK_PIXBUF_ERROR,
+                                             GDK_PIXBUF_ERROR_INSUFFICIENT_MEMORY,
+                                             _("Not enough memory to composite a frame in GIF file"));
+                                return -2;
+                        }
+                    
+                        x = context->frame->x_offset;
+                        y = context->frame->y_offset;
+                        w = gdk_pixbuf_get_width (context->frame->pixbuf);
+                        h = gdk_pixbuf_get_height (context->frame->pixbuf);
+                        if (clip_frame (context, &x, &y, &w, &h))
+                                gdk_pixbuf_copy_area (prev_frame->composited,
+                                                      x, y, w, h,
+                                                      context->frame->pixbuf,
+                                                      0, 0);
                 }
         }
 
@@ -941,36 +1059,29 @@ gif_get_lzw (GifContext *context)
         
 	if (bound_flag && context->update_func) {
 		if (lower_bound <= upper_bound && first_pass == context->draw_pass) {
-			(* context->update_func)
-				(context->frame->pixbuf,
-				 0, lower_bound,
-				 gdk_pixbuf_get_width (context->frame->pixbuf),
-				 upper_bound - lower_bound,
-				 context->user_data);
+                        maybe_update (context, 
+                                      context->frame->x_offset,
+                                      context->frame->y_offset + lower_bound,
+                                      gdk_pixbuf_get_width (context->frame->pixbuf),
+                                      upper_bound - lower_bound);
 		} else {
 			if (lower_bound <= upper_bound) {
-				(* context->update_func)
-					(context->frame->pixbuf,
-					 context->frame->x_offset,
-                                         context->frame->y_offset,
-					 gdk_pixbuf_get_width (context->frame->pixbuf),
-					 gdk_pixbuf_get_height (context->frame->pixbuf),
-					 context->user_data);
+                                maybe_update (context,
+                                              context->frame->x_offset,
+                                              context->frame->y_offset,
+                                              gdk_pixbuf_get_width (context->frame->pixbuf),
+                                              gdk_pixbuf_get_height (context->frame->pixbuf));
 			} else {
-				(* context->update_func)
-					(context->frame->pixbuf,
-					 context->frame->x_offset,
-                                         context->frame->y_offset,
-					 gdk_pixbuf_get_width (context->frame->pixbuf),
-					 upper_bound,
-					 context->user_data);
-				(* context->update_func)
-					(context->frame->pixbuf,
-					 context->frame->x_offset,
-                                         lower_bound + context->frame->y_offset,
-					 gdk_pixbuf_get_width (context->frame->pixbuf),
-					 gdk_pixbuf_get_height (context->frame->pixbuf),
-					 context->user_data);
+                                maybe_update (context,
+                                              context->frame->x_offset,
+                                              context->frame->y_offset,
+                                              gdk_pixbuf_get_width (context->frame->pixbuf),
+                                              upper_bound);
+                                maybe_update (context,
+                                              context->frame->x_offset,
+                                              context->frame->y_offset + lower_bound,
+                                              gdk_pixbuf_get_width (context->frame->pixbuf),
+                                              gdk_pixbuf_get_height (context->frame->pixbuf) - lower_bound);
 			}
 		}
 	}
@@ -981,6 +1092,9 @@ gif_get_lzw (GifContext *context)
                  */
 		context->frame = NULL;
                 context->frame_cmap_active = FALSE;
+
+                if (context->stop_after_first_frame)
+                        context->state =  GIF_DONE;
 	}
 	
 	return v;
@@ -1099,7 +1213,10 @@ gif_init (GifContext *context)
         context->animation->bg_red = 0;
         context->animation->bg_green = 0;
         context->animation->bg_blue = 0;
-        
+
+        context->animation->width = context->width;
+        context->animation->height = context->height;
+
 	if (context->has_global_cmap) {
 		gif_set_get_colormap (context);
 	} else {
@@ -1135,53 +1252,22 @@ gif_get_frame_info (GifContext *context)
 	context->x_offset = LM_to_uint (buf[0], buf[1]);
 	context->y_offset = LM_to_uint (buf[2], buf[3]);
 
-        if ((context->frame_height == 0) || (context->frame_len == 0)) {
-		context->state = GIF_DONE;
-
-                g_set_error (context->error,
-                             GDK_PIXBUF_ERROR,
-                             GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
-                             _("GIF image contained a frame with height or width 0."));
-                
-		return -2;
-        }
-            
-	if (((context->frame_height + context->y_offset) > context->height) ||
-            ((context->frame_len + context->x_offset) > context->width)) {
-		/* All frames must fit in the image bounds */
-		context->state = GIF_DONE;
-
-                g_set_error (context->error,
-                             GDK_PIXBUF_ERROR,
-                             GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
-                             _("GIF image contained a frame appearing outside the image bounds."));
-                
-		return -2;
-	}
-        
 	if (context->animation->frames == NULL &&
             context->gif89.disposal == 3) {
                 /* First frame can't have "revert to previous" as its
-                 * dispose mode.
+                 * dispose mode. Silently use "retain" instead.
                  */
-                
-		context->state = GIF_DONE;
-
-                g_set_error (context->error,
-                             GDK_PIXBUF_ERROR,
-                             GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
-                             _("First frame of GIF image had 'revert to previous' as its disposal mode."));
-                
-		return -2;
+                context->gif89.disposal = 0;
 	}
 
+	context->frame_interlace = BitSet (buf[8], INTERLACE);
+
 #ifdef DUMP_IMAGE_DETAILS
-        g_print (">width: %d height: %d xoffset: %d yoffset: %d disposal: %d delay: %d transparent: %d\n",
+        g_print (">width: %d height: %d xoffset: %d yoffset: %d disposal: %d delay: %d transparent: %d interlace: %d\n",
                  context->frame_len, context->frame_height, context->x_offset, context->y_offset,
-                 context->gif89.disposal, context->gif89.delay_time, context->gif89.transparent);
+                 context->gif89.disposal, context->gif89.delay_time, context->gif89.transparent, context->frame_interlace);
 #endif
         
-	context->frame_interlace = BitSet (buf[8], INTERLACE);
 	if (BitSet (buf[8], LOCALCOLORMAP)) {
 
 #ifdef DUMP_IMAGE_DETAILS
@@ -1232,7 +1318,7 @@ gif_get_next_step (GifContext *context)
 		}
 
 		if (c == '!') {
-			/* Check the extention */
+			/* Check the extension */
 			gif_set_get_extension (context);
 			return 0;
 		}
@@ -1280,7 +1366,7 @@ gif_main_loop (GifContext *context)
 			retval = gif_get_frame_info (context);
 			break;
 
-		case GIF_GET_EXTENTION:
+		case GIF_GET_EXTENSION:
                         LOG("get_extension\n");
 			retval = gif_get_extension (context);
 			if (retval == 0)
@@ -1336,7 +1422,7 @@ new_context (void)
 
         memset (context, 0, sizeof (GifContext));
         
-        context->animation = g_object_new (GDK_TYPE_PIXBUF_GIF_ANIM, NULL);        
+        context->animation = g_object_new (GDK_TYPE_PIXBUF_GIF_ANIM, NULL);
 	context->frame = NULL;
 	context->file = NULL;
 	context->state = GIF_START;
@@ -1349,6 +1435,9 @@ new_context (void)
 	context->gif89.delay_time = -1;
 	context->gif89.input_flag = -1;
 	context->gif89.disposal = -1;
+        context->animation->loop = 1;
+        context->in_loop_extension = FALSE;
+        context->stop_after_first_frame = FALSE;
 
 	return context;
 }
@@ -1373,7 +1462,8 @@ gdk_pixbuf__gif_image_load (FILE *file, GError **error)
         
 	context->file = file;
         context->error = error;
-        
+        context->stop_after_first_frame = TRUE;
+
 	if (gif_main_loop (context) == -1 || context->animation->frames == NULL) {
                 if (context->error && *(context->error) == NULL)
                         g_set_error (context->error,
@@ -1396,9 +1486,9 @@ gdk_pixbuf__gif_image_load (FILE *file, GError **error)
 }
 
 static gpointer
-gdk_pixbuf__gif_image_begin_load (ModuleSizeFunc size_func,
-                                  ModulePreparedNotifyFunc prepare_func,
-				  ModuleUpdatedNotifyFunc update_func,
+gdk_pixbuf__gif_image_begin_load (GdkPixbufModuleSizeFunc size_func,
+                                  GdkPixbufModulePreparedFunc prepare_func,
+				  GdkPixbufModuleUpdatedFunc update_func,
 				  gpointer user_data,
                                   GError **error)
 {
@@ -1495,7 +1585,7 @@ gdk_pixbuf__gif_image_load_increment (gpointer data,
 		/* prepare for the next image_load_increment */
 		if (context->buf == buf) {
 			g_assert (context->size == size);
-			context->buf = (guchar *)g_new (guchar, context->amount_needed + (context->size - context->ptr));
+			context->buf = g_new (guchar, context->amount_needed + (context->size - context->ptr));
 			memcpy (context->buf, buf + context->ptr, context->size - context->ptr);
 		} else {
 			/* copy the left overs to the begining of the buffer */
@@ -1560,11 +1650,36 @@ gdk_pixbuf__gif_image_load_animation (FILE *file,
 }
 
 void
-gdk_pixbuf__gif_fill_vtable (GdkPixbufModule *module)
+MODULE_ENTRY (gif, fill_vtable) (GdkPixbufModule *module)
 {
-  module->load = gdk_pixbuf__gif_image_load;
-  module->begin_load = gdk_pixbuf__gif_image_begin_load;
-  module->stop_load = gdk_pixbuf__gif_image_stop_load;
-  module->load_increment = gdk_pixbuf__gif_image_load_increment;
-  module->load_animation = gdk_pixbuf__gif_image_load_animation;
+        module->load = gdk_pixbuf__gif_image_load;
+        module->begin_load = gdk_pixbuf__gif_image_begin_load;
+        module->stop_load = gdk_pixbuf__gif_image_stop_load;
+        module->load_increment = gdk_pixbuf__gif_image_load_increment;
+        module->load_animation = gdk_pixbuf__gif_image_load_animation;
+}
+
+void
+MODULE_ENTRY (gif, fill_info) (GdkPixbufFormat *info)
+{
+        static GdkPixbufModulePattern signature[] = {
+                { "GIF8", NULL, 100 },
+                { NULL, NULL, 0 }
+        };
+	static gchar * mime_types[] = {
+		"image/gif",
+		NULL
+	};
+	static gchar * extensions[] = {
+		"gif",
+		NULL
+	};
+
+	info->name = "gif";
+        info->signature = signature;
+	info->description = N_("The GIF image format");
+	info->mime_types = mime_types;
+	info->extensions = extensions;
+	info->flags = GDK_PIXBUF_FORMAT_THREADSAFE;
+	info->license = "LGPL";
 }
