@@ -92,6 +92,15 @@ typedef enum {
 
 GQuark gtk_icon_theme_error_quark (void);
 
+#ifdef G_OS_WIN32
+/* Reserve old name for DLL ABI backward compatibility */
+#define gtk_icon_theme_set_search_path gtk_icon_theme_set_search_path_utf8
+#define gtk_icon_theme_get_search_path gtk_icon_theme_get_search_path_utf8
+#define gtk_icon_theme_append_search_path gtk_icon_theme_append_search_path_utf8
+#define gtk_icon_theme_prepend_search_path gtk_icon_theme_prepend_search_path_utf8
+#define gtk_icon_info_get_filename gtk_icon_info_get_filename_utf8
+#endif
+
 GType         gtk_icon_theme_get_type              (void) G_GNUC_CONST;
 
 GtkIconTheme *gtk_icon_theme_new                   (void);
@@ -158,6 +167,10 @@ gboolean              gtk_icon_info_get_attach_points (GtkIconInfo   *icon_info,
 						       gint          *n_points);
 G_CONST_RETURN gchar *gtk_icon_info_get_display_name  (GtkIconInfo   *icon_info);
 
+/* Non-public methods */
+void _gtk_icon_theme_check_reload                     (GdkDisplay *display);
+void _gtk_icon_theme_ensure_builtin_cache             (void);
+
 G_END_DECLS
 
 #endif /* __GTK_ICON_THEME_H__ */
